@@ -9,7 +9,7 @@
  *
  *     E_n(x) = int_x^infty exp(-t)/t^n dt.
  *
- *  Copyright (C) 2016 Vincent Goulet
+ *  Copyright (C) 2016-2026 Vincent Goulet
  *
  *  The code in part IMPLEMENTATION is derived from the GNU Scientific
  *  Library (GSL) v2.2.1 <https://www.gnu.org/software/gsl/>
@@ -19,9 +19,9 @@
  *
  *  The code in part R TO C INTERFACE is derived from R source code.
  *
+ *  Copyright (C) 1998--2025 The R Core Team
+ *  Copyright (C) 2003--2023 The R Foundation
  *  Copyright (C) 1995--1997 Robert Gentleman and Ross Ihaka
- *  Copyright (C) 1998--2016 The R Core Team.
- *  Copyright (C) 2003--2016 The R Foundation
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU General Public License as
@@ -49,6 +49,9 @@
 #include "locale.h"
 #include "expint.h"
 
+/* Prototypes of auxiliary functions */
+static SEXP expint1_1(SEXP, SEXP, double (*f)(double, int));
+static SEXP expint2_1(SEXP, SEXP, SEXP, double (*f)(double, int, int));
 
 /*
  *  IMPLEMENTATION OF THE WORKHORSES
@@ -346,8 +349,8 @@ double expint_E1(double x, int scale)
 	return x;
 #endif
 
-    const double xmaxt = -LOG_DOUBLE_XMIN;      /* XMAXT = -LOG(DOUBLE_XMIN) */
-    const double xmax  = xmaxt - log(xmaxt);    /* XMAX = XMAXT - LOG(XMAXT) */
+    const double xmaxt = -LOG_DBL_MIN;       /* XMAXT = -LOG(DBL_MIN) */
+    const double xmax  = xmaxt - log(xmaxt); /* XMAX = XMAXT - LOG(XMAXT) */
 
     if (x < -xmax && !scale)
     {
@@ -417,7 +420,7 @@ double expint_E2(double x, int scale)
 	return x;
 #endif
 
-    const double xmaxt = -LOG_DOUBLE_XMIN;
+    const double xmaxt = -LOG_DBL_MIN;
     const double xmax  = xmaxt - log(xmaxt);
 
     if (x < -xmax && !scale)
@@ -470,7 +473,7 @@ double expint_E2(double x, int scale)
 
 /* Macro used in expint_En (only) */
 #define CHECK_UNDERFLOW(x)			\
-    if (fabs(x) < DOUBLE_XMIN) {		\
+    if (fabs(x) < DBL_MIN) {			\
         warning(_("underflow in expint_En"));	\
 	return 0.0;				\
     }						\
@@ -533,18 +536,17 @@ double expint_En(double x, int n, int scale)
 
 /* Functions to handle cases with one argument (REAL) and an integer
  * flag */
-static SEXP expint1_1(SEXP sx, SEXP sI, double (*f)())
+static SEXP expint1_1(SEXP sx, SEXP sI, double (*f)(double, int))
 {
     SEXP sy;
-    int i, nx, sxo = OBJECT(sx);
+    R_xlen_t i, nx;
     double xi, *x, *y;
-    int i_1;
     Rboolean naflag = FALSE;
 
     if (!isNumeric(sx))
         error(_("invalid arguments"));
 
-    nx = LENGTH(sx);
+    nx = XLENGTH(sx);
     if (nx == 0)
         return(allocVector(REALSXP, 0));
     PROTECT(sx = coerceVector(sx, REALSXP));
@@ -552,7 +554,7 @@ static SEXP expint1_1(SEXP sx, SEXP sI, double (*f)())
     x = REAL(sx);
     y = REAL(sy);
 
-    i_1 = asInteger(sI);
+    int i_1 = asInteger(sI);
 
     for (i = 0; i < nx; i++)
     {
@@ -571,8 +573,7 @@ static SEXP expint1_1(SEXP sx, SEXP sI, double (*f)())
     if (naflag)
         warning(R_MSG_NA);
 
-    SET_ATTRIB(sy, duplicate(ATTRIB(sx)));
-    SET_OBJECT(sy, sxo);
+    SHALLOW_DUPLICATE_ATTRIB(sy, sx);
     UNPROTECT(2);
 
     return sy;
@@ -601,20 +602,19 @@ SEXP expint_do_expint1(int code, SEXP args)
              i2 = (++i2 == n2) ? 0 : i2,        \
              ++i)
 
-static SEXP expint2_1(SEXP sx, SEXP sa, SEXP sI, double (*f)())
+static SEXP expint2_1(SEXP sx, SEXP sa, SEXP sI, double (*f)(double, int, int))
 {
     SEXP sy;
-    int i, ix, ia, n, nx, na,
-	sxo = OBJECT(sx), sao = OBJECT(sa);
+    R_xlen_t i, ix, ia, n, nx, na;
     double xi, *x, *y;
-    int ai, *a, i_1;
+    int ai, *a;
     Rboolean naflag = FALSE;
 
     if (!isNumeric(sx) || !isNumeric(sa))
         error(_("invalid arguments"));
 
-    nx = LENGTH(sx);
-    na = LENGTH(sa);
+    nx = XLENGTH(sx);
+    na = XLENGTH(sa);
     if ((nx == 0) || (na == 0))
         return(allocVector(REALSXP, 0));
 
@@ -627,7 +627,7 @@ static SEXP expint2_1(SEXP sx, SEXP sa, SEXP sI, double (*f)())
     a = INTEGER(sa);
     y = REAL(sy);
 
-    i_1 = asInteger(sI);
+    int i_1 = asInteger(sI);
 
     mod_iterate2(nx, na, ix, ia)
     {
@@ -644,7 +644,7 @@ static SEXP expint2_1(SEXP sx, SEXP sa, SEXP sI, double (*f)())
 	    else if (ai == 2)
 		y[i] = expint_E2(xi, i_1);
 	    else
-		y[i] = expint_En(xi, ai, i_1);
+		y[i] = f(xi, ai, i_1);
 	    if (ISNAN(y[i])) naflag = TRUE;
         }
     }
@@ -653,15 +653,9 @@ static SEXP expint2_1(SEXP sx, SEXP sa, SEXP sI, double (*f)())
         warning(R_MSG_NA);
 
     if (n == nx)
-    {
-        SET_ATTRIB(sy, duplicate(ATTRIB(sx)));
-        SET_OBJECT(sy, sxo);
-    }
+        SHALLOW_DUPLICATE_ATTRIB(sy, sx);
     else if (n == na)
-    {
-        SET_ATTRIB(sy, duplicate(ATTRIB(sa)));
-        SET_OBJECT(sy, sao);
-    }
+        SHALLOW_DUPLICATE_ATTRIB(sy, sa);
 
     UNPROTECT(3);
 
@@ -701,7 +695,7 @@ static expint_tab_struct expint_tab[] = {
 /* Function called by .External() */
 SEXP expint_do_expint(SEXP args)
 {
-    int i;
+    R_xlen_t i;
     const char *name;
 
     /* Extract type of exponential integral */
